/*
 * SWTComponentPeer.java
 *
 * Copyright (c) 2007 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.swtawt;

import java.awt.AWTEvent;
import java.awt.BufferCapabilities;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.ScrollPane;
import java.awt.BufferCapabilities.FlipContents;
import java.awt.event.ComponentEvent;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.PaintEvent;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ImageObserver;
import java.awt.image.ImageProducer;
import java.awt.image.VolatileImage;
import java.awt.peer.ComponentPeer;
import java.awt.peer.ContainerPeer;
import java.awt.peer.LightweightPeer;
import java.lang.reflect.Field;
import java.util.concurrent.Callable;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.PaletteData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;

import sun.awt.CausedFocusEvent;

/**
 * @author Emmanuel Puybaret
 */
public abstract class SWTComponentPeer extends SWTObjectPeer implements ComponentPeer {
  private final static Field AWT_COMPONENT_X_FIELD;
  private final static Field AWT_COMPONENT_Y_FIELD;
  private final static Field AWT_COMPONENT_WIDTH_FIELD;
  private final static Field AWT_COMPONENT_HEIGHT_FIELD;

  private int     mousePressedClickCount = 0;
  private long    mousePressedTime       = 0;
  private boolean mouseMovedSinceLastMousePressed = false;
  private boolean layouting;
  private boolean paintPending;
  private BufferedImage componentImage;
  
  static {
    try {
      AWT_COMPONENT_X_FIELD = Component.class.getDeclaredField("x");
      AWT_COMPONENT_X_FIELD.setAccessible(true);
      AWT_COMPONENT_Y_FIELD = Component.class.getDeclaredField("y");
      AWT_COMPONENT_Y_FIELD.setAccessible(true);
      AWT_COMPONENT_WIDTH_FIELD = Component.class.getDeclaredField("width");
      AWT_COMPONENT_WIDTH_FIELD.setAccessible(true);
      AWT_COMPONENT_HEIGHT_FIELD = Component.class.getDeclaredField("height");
      AWT_COMPONENT_HEIGHT_FIELD.setAccessible(true);
    } catch (SecurityException ex) {
      throw new RuntimeException("Granting access to Component fields is required to post component moved and resized events", ex);
    } catch (NoSuchFieldException ex) {
      throw new RuntimeException(ex);
    }
  }
  
  public SWTComponentPeer(SWTToolkit toolkit, final Component awtComponent) {
    super(toolkit, awtComponent);
    initSWTWidget();
    
    setEnabled(awtComponent.isEnabled());
    setBounds(awtComponent.getX(), awtComponent.getY(), 
        awtComponent.getWidth(), awtComponent.getHeight(), ComponentPeer.SET_BOUNDS);
    
    // TODO Managed better isForegroundSet, isBackgroundSet and isFontSet
    if (awtComponent.isForegroundSet()) {
      Color foreground = awtComponent.getForeground();
      if (foreground != null) {
        setForeground(foreground);
      }
    }
    if (awtComponent.isBackgroundSet()) {
      Color background = awtComponent.getBackground();
      if (background != null) {
        setBackground(background);
      }
    }
    Font font = awtComponent.getFont();
    if (font != null) {
      setFont(font);
    }
  }

  protected void initSWTWidget() {
    super.initSWTWidget();
    
    if (getAWTObject().getParent() instanceof ScrollPane) {
      invokeInSWTDisplayThread(new Callable<Void>() {
        public Void call() {
          ((ScrolledComposite)getSWTParent()).setContent((Control)getSWTWidget());
          return null;
        }
      });
    }
  }
  
  @Override
  protected Component getAWTObject() {
    return (Component)super.getAWTObject();
  }
  
  @Override
  protected Control getSWTWidget() {
    return (Control)super.getSWTWidget();
  }

  /**
   * Returns the SWT composite associated with the parent of the AWT component of this peer.
   */
  @SuppressWarnings("deprecation")
  protected Composite getSWTParent() {
    Container awtParent = getAWTObject().getParent();
    // In case awtParent is a light weight component find the first heavy weight one in hierarchy 
    while (awtParent.getPeer() instanceof LightweightPeer) {
      awtParent = awtParent.getParent();
    }
    if (awtParent == null) {
      return null;
    } else {
      return (Composite)((SWTComponentPeer)awtParent.getPeer()).getSWTWidget();
    }
  }
  
  /**
   * Handles SWT move events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMoveEvent(Event ev) {
    // Update component location and post a COMPONENT_MOVED AWT event 
    EventQueue.invokeLater(new Runnable() {
      public void run() {
        try {
          final Rectangle bounds = getBounds();
          AWT_COMPONENT_X_FIELD.set(getAWTObject(), bounds.x);
          AWT_COMPONENT_Y_FIELD.set(getAWTObject(), bounds.y);
        } catch (IllegalAccessException ex) {
          throw new RuntimeException("Granting write access is required to post component resized events", ex);
        }
      }
    });
    postAWTEvent(new ComponentEvent(getAWTObject(), ComponentEvent.COMPONENT_MOVED));
  }
  
  /**
   * Handles SWT resize events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTResizeEvent(Event ev) {
    // Update component size and post a COMPONENT_RESIZED AWT event 
    EventQueue.invokeLater(new Runnable() {
      public void run() {
        try {
          final Rectangle bounds = getBounds();
          AWT_COMPONENT_WIDTH_FIELD.set(getAWTObject(), bounds.width);
          AWT_COMPONENT_HEIGHT_FIELD.set(getAWTObject(), bounds.height);
        } catch (IllegalAccessException ex) {
          throw new RuntimeException("Granting write access is required to post component resized events", ex);
        }
      }
    });
    postAWTEvent(new ComponentEvent(getAWTObject(), ComponentEvent.COMPONENT_RESIZED));
  }
  
  /**
   * Handles SWT show events.
   * This method is executed in SWT display event loop thread.  
   */
  @Override
  protected void handleSWTShowEvent(Event ev) {
    postAWTEvent(new ComponentEvent(getAWTObject(), ComponentEvent.COMPONENT_SHOWN));
  }
  
  /**
   * Handles SWT hide events.
   * This method is executed in SWT display event loop thread.  
   */
  @Override
  protected void handleSWTHideEvent(Event ev) {
    postAWTEvent(new ComponentEvent(getAWTObject(), ComponentEvent.COMPONENT_HIDDEN));
  }
  
  /**
   * Handles SWT activate events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTActivateEvent(Event ev) {
  }
  
  /**
   * Handles SWT deactivate events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTDeactivateEvent(Event ev) {
  }

  /**
   * Handles SWT focus in events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTFocusInEvent(Event ev) {
    postAWTEvent(new FocusEvent(getAWTObject(), FocusEvent.FOCUS_GAINED));
  }
  
  /**
   * Handles SWT focus out events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTFocusOutEvent(Event ev) {
    postAWTEvent(new FocusEvent(getAWTObject(), FocusEvent.FOCUS_LOST));        
  }
  
  /**
   * Handles SWT mouse down events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseDownEvent(Event ev) {
    if (this.mouseMovedSinceLastMousePressed
        || (ev.time - this.mousePressedTime) > getToolkit().getSWTDisplay().getDoubleClickTime()) {
      this.mousePressedClickCount = 0;
    }
    this.mouseMovedSinceLastMousePressed = false; 
    this.mousePressedTime = ev.time; 
    this.mousePressedClickCount++; 
    postAWTEvent(new MouseEvent(getAWTObject(), MouseEvent.MOUSE_PRESSED, ev.time, 
        getModifiers(ev.stateMask), ev.x, ev.y, this.mousePressedClickCount, 
        ev.button == 3 || (getToolkit().isMacOSX() && ev.button == 1 && ev.stateMask == SWT.CTRL), 
        getButton(ev.button)));        
  }
  
  /**
   * Handles SWT double click events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseDoubleClickEvent(Event ev) {
    // Nothing to do, double click are handled my mouse down 
  }
  
  /**
   * Handles SWT mouse up events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseUpEvent(Event ev) {
    postAWTEvent(new MouseEvent(getAWTObject(), MouseEvent.MOUSE_RELEASED, ev.time, 
        getModifiers(ev.stateMask), ev.x, ev.y, 
        this.mousePressedClickCount, false, getButton(ev.button)));
    if (!this.mouseMovedSinceLastMousePressed) {
      postAWTEvent(new MouseEvent(getAWTObject(), MouseEvent.MOUSE_CLICKED, ev.time, 
          getModifiers(ev.stateMask), ev.x, ev.y, 
          this.mousePressedClickCount, false, getButton(ev.button)));
    } 
  }
  
  /**
   * Handles SWT mouse move events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseMoveEvent(Event ev) {
    this.mouseMovedSinceLastMousePressed = true;
    boolean buttonDepressed = (ev.stateMask & (SWT.BUTTON1 | SWT.BUTTON2 | SWT.BUTTON3)) != 0;
    postAWTEvent(new MouseEvent(getAWTObject(), 
        buttonDepressed ? MouseEvent.MOUSE_DRAGGED : MouseEvent.MOUSE_MOVED, 
        ev.time, getModifiers(ev.stateMask), ev.x, ev.y, 
        this.mousePressedClickCount, false, getButton(ev.button)));
  }
  
  /**
   * Handles SWT mouse wheel events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseWheelEvent(Event ev) {
    postAWTEvent(new MouseWheelEvent(getAWTObject(), MouseEvent.MOUSE_WHEEL, ev.time, 
        getModifiers(ev.stateMask), ev.x, ev.y, 
        this.mousePressedClickCount, false,  
        MouseWheelEvent.WHEEL_UNIT_SCROLL, ev.count, ev.count));
  }
  
  /**
   * Handles SWT mouse enter events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseEnterEvent(Event ev) {
    postAWTEvent(new MouseEvent(getAWTObject(), MouseEvent.MOUSE_ENTERED, ev.time, 
        getModifiers(ev.stateMask), ev.x, ev.y, 1, false, getButton(ev.button)));
  }
  
  /**
   * Handles SWT mouse enter events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseExitEvent(Event ev) {
    this.mouseMovedSinceLastMousePressed = true; 
    postAWTEvent(new MouseEvent(getAWTObject(), MouseEvent.MOUSE_EXITED, ev.time, 
        getModifiers(ev.stateMask), ev.x, ev.y, 
        this.mousePressedClickCount, false, getButton(ev.button)));
  }
  
  /**
   * Handles SWT mouse hover events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTMouseHoverEvent(Event ev) {
    this.mouseMovedSinceLastMousePressed = true; 
    postAWTEvent(new MouseEvent(getAWTObject(), MouseEvent.MOUSE_MOVED, ev.time, 
        getModifiers(ev.stateMask), ev.x, ev.y, 
        this.mousePressedClickCount, false, getButton(ev.button)));
  }
  
  /**
   * Handles SWT key down events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTKeyDownEvent(Event ev) {
    postAWTEvent(new KeyEvent(getAWTObject(), KeyEvent.KEY_PRESSED, ev.time, 
        getModifiers(ev.stateMask), getAWTKeyCode(ev.keyCode), ev.character));
  }
  
  /**
   * Handles SWT key up events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTKeyUpEvent(Event ev) {
    postAWTEvent(new KeyEvent(getAWTObject(), KeyEvent.KEY_RELEASED, ev.time, 
        getModifiers(ev.stateMask), getAWTKeyCode(ev.keyCode), ev.character));
  }
  
  /**
   * Handles SWT paint events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTPaintEvent(Event ev) {
    if (!this.paintPending) {
      postAWTEvent(new PaintEvent(getAWTObject(), PaintEvent.PAINT, 
          new Rectangle(ev.x, ev.y, ev.width, ev.height)));
    } 
  }
  
  /**
   * Handles SWT modify events.
   * This method is executed in SWT display event loop thread.  
   */
  protected void handleSWTModifyEvent(Event ev) {
  }
  
  private int getButton(int buttonIndex) {
    switch (buttonIndex) {
      case 1 : return MouseEvent.BUTTON1;
      case 2 : return MouseEvent.BUTTON2;
      case 3 : return MouseEvent.BUTTON3;
      default : return 0;
    }
  }

  public void setBackground(final Color color) {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTWidget().setBackground(SWTColors.getInstance(getToolkit().getSWTDisplay()).getColor(color));
        return null;
      }
    });
  }

  public void setForeground(final Color color) {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTWidget().setForeground(SWTColors.getInstance(getToolkit().getSWTDisplay()).getColor(color));
        return null;
      }
    });
  }

  public void setFont(final Font font) {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTWidget().setFont(SWTFonts.getInstance(getToolkit().getSWTDisplay()).getFont(font));
        return null;
      }
    });
  }

  final public void disable() {
    setEnabled(false);
  }

  final public void enable() {
    setEnabled(true);
  }

  public void setEnabled(final boolean enabled) {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTWidget().setEnabled(enabled);
        return null;
      }
    });
  }

  public void hide() {
    setVisible(false);
  }
  
  public void show() {
    setVisible(true);
  }
  
  public void setVisible(final boolean visible) {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTWidget().setVisible(visible);
        return null;
      }
    });
  }

  public Point getLocationOnScreen() {
    return invokeInSWTDisplayThread(new Callable<Point>() {
      public Point call() {
        org.eclipse.swt.graphics.Point location = getSWTWidget().getLocation();
        location = Display.getDefault().map(getSWTWidget(), null, location);
        return new Point (location.x, location.y);
      }
    });
  }

  public Rectangle getBounds() {
    return invokeInSWTDisplayThread(new Callable<Rectangle>() {
      public Rectangle call() {
        org.eclipse.swt.graphics.Rectangle bounds = getSWTWidget().getBounds();
        return new Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);
      }
    });
  }


  public Dimension preferredSize() {
    return getPreferredSize();
  }

  public Dimension getPreferredSize() {
    return invokeInSWTDisplayThread(new Callable<Dimension>() {
      public Dimension call() {
        org.eclipse.swt.graphics.Point size = getSWTWidget().computeSize(SWT.DEFAULT, SWT.DEFAULT);
        return new Dimension(size.x, size.y);
      }
    });
  }

  public Dimension minimumSize() {
    return getMinimumSize();
  }

  public Dimension getMinimumSize() {
    // TODO Find better ?
    return getPreferredSize();
  }

  public void reshape(final int x, final int y, final int width, final int height) {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTWidget().setBounds(x, y, width, height);
        return null;
      }
    });
  }

  public void setBounds(int x, int y, int width, int height, int op) {
    reshape(x, y, width, height);
  }

  public boolean canDetermineObscurity() {
    // TODO Can obscurity be determined ?
    return false;
  }

  public boolean isObscured() {
    return false;
  }

  public Image createImage(ImageProducer producer) {
    return getToolkit().createImage(producer);
  }

  public Image createImage(int width, int height) {
    return new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
  }

  public int checkImage(Image image, int width, int height, ImageObserver observer) {
    return getToolkit().checkImage(image, width, height, observer);
  }

  public boolean prepareImage(Image image, int width, int height, ImageObserver observer) {
    return getToolkit().prepareImage(image, width, height, observer);
  }

  public VolatileImage createVolatileImage(int width, int height) {
    // TODO Manage volatile images
    return null;
  }

  public void flip(FlipContents flipAction) {
    // TODO Manage volatile images
  }

  public Image getBackBuffer() {
    // TODO Manage volatile images
    return null;
  }

  public void createBuffers(int numBuffers, BufferCapabilities caps) {
    // TODO Manage volatile images
  }

  public void destroyBuffers() {
    // TODO Manage volatile images
  }

  public ColorModel getColorModel() {
    return getGraphicsConfiguration().getColorModel();
  }

  public FontMetrics getFontMetrics(Font font) {
    return getToolkit().getFontMetrics(font);
  }

  public GraphicsConfiguration getGraphicsConfiguration() {
    return getAWTObject().getGraphicsConfiguration();
  }

  public Graphics getGraphics() {
    // Return a Graphics bound to a buffered image to benefit easily of Sun Graphics2D features 
    if (this.componentImage == null
        || this.componentImage.getWidth() != getAWTObject().getWidth()
        || this.componentImage.getHeight() != getAWTObject().getHeight()) {
      this.componentImage = new BufferedImage(
          getAWTObject().getWidth(), getAWTObject().getHeight(), BufferedImage.TYPE_INT_ARGB);
    }
    return this.componentImage.getGraphics();
  }

  public void print(Graphics g) {
    getAWTObject().print(g);
  }

  public void paint(Graphics g) {
    getAWTObject().paint(g);
  }

  public void repaint(long tm, int x, int y, int width, int height) {
  }

  public boolean isPaintPending() {
    return this.paintPending;
  }

  public void coalescePaintEvent(PaintEvent e) {
    // TODO Manage coalesce paint events ?
  }

  public void handleEvent(AWTEvent ev) {
    switch (ev.getID()) {
      case PaintEvent.PAINT:
      case PaintEvent.UPDATE:
        // Skip all painting while layouting
        if (!this.layouting && !this.paintPending) {
          this.paintPending = true;
          Graphics graphics = getGraphics();
          paint(graphics);
          graphics.dispose();
          
          invokeInSWTDisplayThread(new Callable<Void>() {
            public Void call() {
              // TODO optimize SWT image creation (allocate ImageData first ?)
              int [] pixels = componentImage.getRGB(0, 0, componentImage.getWidth(), componentImage.getHeight(), 
                  null, 0, componentImage.getWidth());
              // Copy componentImage pixels to SWT ImageData
              byte [] data = new byte [pixels.length * 3];
              byte [] alphas = new byte [pixels.length];
              for (int i = 0, j = 0; i < pixels.length; i++) {
                alphas [i] = (byte)((pixels [i] >> 24) & 0xFF);
                data [j++] = (byte)((pixels [i] >> 16) & 0xFF); // Red
                data [j++] = (byte)((pixels [i] >> 8) & 0xFF);  // Green
                data [j++] = (byte)((pixels [i] >> 0) & 0xFF);  // Blue
              }
              PaletteData palette = new PaletteData(0xFF0000, 0x00FF00, 0x0000FF);
              ImageData imageData = new ImageData(componentImage.getWidth(), componentImage.getHeight(), 
                  24, palette, componentImage.getWidth(), data);
              imageData.alphaData = alphas;

              // Draw image in component
              GC gc = new GC(getSWTWidget());
              org.eclipse.swt.graphics.Image image = new org.eclipse.swt.graphics.Image(getToolkit().getSWTDisplay(), imageData);
              gc.drawImage(image, 0, 0);
              image.dispose();
              gc.dispose();
              
              paintPending = false;
              return null;
            }
          });
        } 
        return;
      default:
        break;
    }
    // TODO Repost some events like in awt_Component.cpp::_NativeHandleEvent ?
  }

  public void beginLayout() {
    this.layouting = true;
  }

  public void endLayout() {
    if(!getAWTObject().getIgnoreRepaint() && !this.paintPending) {
      postAWTEvent(new PaintEvent(getAWTObject(), PaintEvent.PAINT, new Rectangle()));
    }
    this.layouting = false;
  }
  
  public void layout() {
  }

  public boolean isReparentSupported() {
    return false;
  }

  public void reparent(ContainerPeer newContainer) {
    // getSWTWidget().setParent(((SWTContainerPeer)newContainer).getSWTParent());
  }

  public boolean handlesWheelScrolling() {
    return false;
  }

  public boolean isFocusable() {
    return false;
  }

  public boolean requestFocus(Component lightweightChild,
                              boolean temporary,
                              boolean focusedWindowChangeAllowed,
                              long time) {
    return invokeInSWTDisplayThread(new Callable<Boolean>() {
      public Boolean call() {
        return getSWTWidget().setFocus();
      }
    });
  }

  public boolean requestFocus(Component lightweightChild,
                              boolean temporary,
                              boolean focusedWindowChangeAllowed,
                              long time, CausedFocusEvent.Cause cause) {
    return requestFocus(lightweightChild, temporary, focusedWindowChangeAllowed, time);
  }
  
  public void updateCursorImmediately() {
    // TODO Manage custom cursors
    /* TODO Find why this gets to a deadlock ???
    Cursor awtCursor = getAWTObject().getCursor(); 
    final int swtCursorType = getSWTCursorType(awtCursor);
    if (swtCursorType != 0) {
      invokeInSWTDisplayThread(new Callable<Void>() {
        public Void call() {          
          org.eclipse.swt.graphics.Cursor oldCursor = getSWTWidget().getCursor(); 
          getSWTWidget().setCursor(
              new org.eclipse.swt.graphics.Cursor(getToolkit().getSWTDisplay(), swtCursorType));
          if (oldCursor != null) {
            oldCursor.dispose();
          }
          return null;
        }
      });
    }
    */
  }

  private int getSWTCursorType(Cursor awtCursor) {
    switch (awtCursor.getType()) {
      case Cursor.DEFAULT_CURSOR:
        return SWT.CURSOR_ARROW;
      case Cursor.CROSSHAIR_CURSOR:
        return SWT.CURSOR_CROSS;
      case Cursor.TEXT_CURSOR:
        return SWT.CURSOR_IBEAM;
      case Cursor.WAIT_CURSOR:
        return SWT.CURSOR_WAIT;
      case Cursor.SW_RESIZE_CURSOR:
        return SWT.CURSOR_SIZESW;
      case Cursor.SE_RESIZE_CURSOR:
        return SWT.CURSOR_SIZESE;
      case Cursor.NW_RESIZE_CURSOR:
        return SWT.CURSOR_SIZENW;
      case Cursor.NE_RESIZE_CURSOR:
        return SWT.CURSOR_SIZENE;
      case Cursor.N_RESIZE_CURSOR:
        return SWT.CURSOR_SIZEN;
      case Cursor.S_RESIZE_CURSOR:
        return SWT.CURSOR_SIZES;
      case Cursor.W_RESIZE_CURSOR:
        return SWT.CURSOR_SIZEW;
      case Cursor.E_RESIZE_CURSOR:
        return SWT.CURSOR_SIZEE;
      case Cursor.HAND_CURSOR:
        return SWT.CURSOR_HAND;
      case Cursor.MOVE_CURSOR:
        return SWT.CURSOR_SIZEALL;
      default:
        return 0;
    }
  }
}
