/*
 * SWTToolkit.java
 *
 * Copyright (c) 2007 Emmanuel PUYBARET / eTeks <info@eteks.com>. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.swtawt;

import java.awt.Button;
import java.awt.Canvas;
import java.awt.Checkbox;
import java.awt.CheckboxMenuItem;
import java.awt.Choice;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.FileDialog;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Frame;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.HeadlessException;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Label;
import java.awt.List;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.Panel;
import java.awt.Point;
import java.awt.PopupMenu;
import java.awt.PrintJob;
import java.awt.Robot;
import java.awt.ScrollPane;
import java.awt.Scrollbar;
import java.awt.SystemTray;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.TrayIcon;
import java.awt.Window;
import java.awt.datatransfer.Clipboard;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.InvalidDnDOperationException;
import java.awt.dnd.peer.DragSourceContextPeer;
import java.awt.event.KeyEvent;
import java.awt.font.TextAttribute;
import java.awt.im.InputMethodHighlight;
import java.awt.im.spi.InputMethodDescriptor;
import java.awt.image.ColorModel;
import java.awt.peer.ButtonPeer;
import java.awt.peer.CanvasPeer;
import java.awt.peer.CheckboxMenuItemPeer;
import java.awt.peer.CheckboxPeer;
import java.awt.peer.ChoicePeer;
import java.awt.peer.DesktopPeer;
import java.awt.peer.DialogPeer;
import java.awt.peer.FileDialogPeer;
import java.awt.peer.FontPeer;
import java.awt.peer.FramePeer;
import java.awt.peer.LabelPeer;
import java.awt.peer.ListPeer;
import java.awt.peer.MenuBarPeer;
import java.awt.peer.MenuItemPeer;
import java.awt.peer.MenuPeer;
import java.awt.peer.PanelPeer;
import java.awt.peer.PopupMenuPeer;
import java.awt.peer.RobotPeer;
import java.awt.peer.ScrollPanePeer;
import java.awt.peer.ScrollbarPeer;
import java.awt.peer.SystemTrayPeer;
import java.awt.peer.TextAreaPeer;
import java.awt.peer.TextFieldPeer;
import java.awt.peer.TrayIconPeer;
import java.awt.peer.WindowPeer;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.concurrent.Callable;
import java.util.concurrent.CountDownLatch;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Widget;

import sun.awt.SunToolkit;

/**
 * @author Emmanuel Puybaret
 */
public class SWTToolkit extends SunToolkit {
  private boolean dynamicLayout;

  private Display swtDisplay;
  private GC swtCachedGc;
  private Map<Widget, SWTObjectPeer> objectPeers;

  public SWTToolkit() {    
    this.objectPeers = new WeakHashMap<Widget, SWTObjectPeer>();
  }
  
  /**
   * Launches SWT display Event loop in a new thread 
   * and waits until this new thread is started.
   */
  private void launchSWTDisplayThread() {
    final CountDownLatch latch = new CountDownLatch(1);
    new Thread("SWT Event loop") { 
      private long noShellTime = System.currentTimeMillis();
      
      @Override
      public void run() {
        // Create SWT display and run SWT Event loop in the same thread
        swtDisplay = new Display();
        // Add to SWT display listeners that tracks all input events
        addSWTListeners();                
        latch.countDown();
        while (true) {
          if (!swtDisplay.readAndDispatch()) {            
            // If SWT display still contains some frame sleep until next event
            if (swtDisplay.getShells().length != 0) {
              swtDisplay.sleep();
              noShellTime = System.currentTimeMillis();
            } else {
              // Otherwise sleep 10 ms
              try {                
                Thread.sleep(10);
              } catch (InterruptedException ex) {
              }
              // And stop the SWT event loop if no SWT event happened during the last 2 seconds
              if (System.currentTimeMillis() - noShellTime > 2000) {
                swtDisplay.dispose();
                objectPeers.clear();
                swtDisplay = null;
                break;
              }
            }
          }
        }
      }
    }.start();
    
    try {
      latch.await();
    } catch (InterruptedException ex) {
      throw new RuntimeException("Couldn't start SWT Event loop");
    }
  }

  /**
   * Adds to SWT display event listeners to track events on SWT components and 
   * repost them in AWT Event Dispatch Thread. This avoids to add on each SWT control
   * their event listeners individually.
   */
  private void addSWTListeners() {
    this.swtDisplay.addFilter(SWT.Move, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMoveEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Resize, new Listener() {
      public void handleEvent(Event ev) {
        SWTComponentPeer componentPeer = (SWTComponentPeer)objectPeers.get(ev.widget);
        // Ignore events posted by widgets that are not yet registered in toolkit
        // (like scroll pane) 
        if (componentPeer != null) {
          componentPeer.handleSWTResizeEvent(ev);
        }
      }
    });
    this.swtDisplay.addFilter(SWT.Show, new Listener() {
      public void handleEvent(Event ev) {
        SWTObjectPeer objectPeer = objectPeers.get(ev.widget);
        // Ignore events posted by widgets that are not registered in toolkit
        // (like FileDialog shell) 
        if (objectPeer != null) {
          objectPeer.handleSWTShowEvent(ev);
        }        
      }
    });
    this.swtDisplay.addFilter(SWT.Hide, new Listener() {
      public void handleEvent(Event ev) {
        SWTObjectPeer objectPeer = objectPeers.get(ev.widget);
        // Ignore events posted by widgets that are not yet registered in toolkit
        // (like scroll pane scroll bars) 
        if (objectPeer != null) {
          objectPeer.handleSWTHideEvent(ev);
        }
      }
    });
    this.swtDisplay.addFilter(SWT.Activate, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTActivateEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Deactivate, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTDeactivateEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Close, new Listener() {
      public void handleEvent(Event ev) {
        if (ev.widget != null) {
          ((SWTWindowPeer)objectPeers.get(ev.widget)).handleSWTCloseEvent(ev);
        }
      }
    });
    this.swtDisplay.addFilter(SWT.Iconify, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTWindowPeer)objectPeers.get(ev.widget)).handleSWTIconifyEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Deiconify, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTWindowPeer)objectPeers.get(ev.widget)).handleSWTDeiconifyEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.FocusIn, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTFocusInEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.FocusOut, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTFocusOutEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseDown, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseDownEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseDoubleClick, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseDoubleClickEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseUp, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseUpEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseMove, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseMoveEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseWheel, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseWheelEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseEnter, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseEnterEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseExit, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseExitEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MouseHover, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTMouseHoverEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.KeyDown, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTKeyDownEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.KeyUp, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTKeyUpEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Paint, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTPaintEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Modify, new Listener() {
      public void handleEvent(Event ev) {
        ((SWTComponentPeer)objectPeers.get(ev.widget)).handleSWTModifyEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.Selection, new Listener() {
      public void handleEvent(Event ev) {
        SWTObjectPeer objectPeer = objectPeers.get(ev.widget);
        // Ignore events posted by widgets that are not registered in toolkit
        // (like list, text... scroll bars) 
        if (objectPeer != null) {
          objectPeer.handleSWTSelectionEvent(ev);
        }
      }
    });
    this.swtDisplay.addFilter(SWT.DefaultSelection, new Listener() {
      public void handleEvent(Event ev) {
        objectPeers.get(ev.widget).handleSWTDefaultSelectionEvent(ev);
      }
    });
    this.swtDisplay.addFilter(SWT.MenuDetect, new Listener() {
      public void handleEvent(Event ev) {
        objectPeers.get(ev.widget).handleSWTMenuDetectEvent(ev);
      }
    });
  }
  
  /**
   * Invokes the <code>call</code> method of <code>callable</code> in 
   * SWT display event loop thread and returns the value returned by <code>call</code>.
   */
  @SuppressWarnings("unchecked")
  <V> V invokeInSWTDisplayThread(final Callable<V> callable) {
    final Exception [] exception = new Exception [1];  
    final V [] returnedValue = (V [])new Object [1];
    
    if (this.swtDisplay == null) {
      // Launch automatically an event loop on SWT display
      launchSWTDisplayThread();
    }
    
    // If we are already in SWT display event loop thread
    if (Thread.currentThread() == this.swtDisplay.getThread()) {
      // Simply invoke call method
      try {
        return callable.call();
      } catch (RuntimeException ex) {
        throw ex; 
      } catch (Exception ex) {
        throw new RuntimeException(ex);
      }
    } else {
      // Otherwise invoke call method synchronously in SWT display event loop thread
      this.swtDisplay.syncExec(new Runnable() {
        public void run() {
          try {
            returnedValue [0] = callable.call();
          } catch (Exception ex) {
            exception [0] = ex;
          }
        }
      });
      // If callable threw an exception report it to current thread
      if (exception [0] != null) {
        if (exception [0] instanceof RuntimeException) {
          throw (RuntimeException)exception [0];
        } else {
          throw new RuntimeException(exception [0]);
        }
      }
      return returnedValue [0];
    }
  }
  
  Display getSWTDisplay() {
    if (this.swtDisplay == null) {
      // Launch automatically an event loop on SWT display
      launchSWTDisplayThread();
    }

    return this.swtDisplay;
  }
  
  public void registerSWTWidget(Widget swtWidget,
                                SWTObjectPeer objectPeer) {
    this.objectPeers.put(swtWidget, objectPeer);
  }

  
  @Override
  public void beep() {
    invokeInSWTDisplayThread(new Callable<Void>() {
      public Void call() {
        getSWTDisplay().beep();
        return null;
      }
    });
  }  

  @Override
  public LabelPeer createLabel(Label target) {
    return new SWTLabelPeer(this, target);
  }

  @Override
  public ButtonPeer createButton(Button target) {
    return new SWTButtonPeer(this, target);
  }

  @Override
  public TextFieldPeer createTextField(TextField target) {
    return new SWTTextFieldPeer(this, target);
  }

  @Override
  public TextAreaPeer createTextArea(TextArea target) {
    return new SWTTextAreaPeer(this, target);
  }

  @Override
  public CheckboxPeer createCheckbox(Checkbox target) {
    return new SWTCheckboxPeer(this, target);
  }

  @Override
  public ChoicePeer createChoice(Choice target) {
    return new SWTChoicePeer(this, target);
  }

  @Override
  public ListPeer createList(List target) {
    return new SWTListPeer(this, target);
  }

  @Override
  public MenuBarPeer createMenuBar(MenuBar target) {
    return new SWTMenuBarPeer(this, target);
  }

  @Override
  public MenuPeer createMenu(Menu target) {
    return new SWTMenuPeer(this, target);
  }

  @Override
  public MenuItemPeer createMenuItem(MenuItem target) {
    return new SWTMenuItemPeer(this, target);
  }

  @Override
  public CheckboxMenuItemPeer createCheckboxMenuItem(CheckboxMenuItem target) {
    return new SWTCheckboxMenuItemPeer(this, target);
  }

  @Override
  public PopupMenuPeer createPopupMenu(PopupMenu target) {
    return new SWTPopupMenuPeer(this, target);
  }

  @Override
  public PanelPeer createPanel(Panel target) {
    return new SWTPanelPeer(this, target);
  }

  @Override
  public ScrollPanePeer createScrollPane(ScrollPane target) {
    return new SWTScrollPanePeer(this, target);
  }

  @Override
  public ScrollbarPeer createScrollbar(Scrollbar target)
  throws HeadlessException {
    return new SWTScrollbarPeer(this, target);
  }
  
  @Override
  public CanvasPeer createCanvas(Canvas target) {
    return new SWTCanvasPeer(this, target);
  }
  
  @Override
  public WindowPeer createWindow(Window target) {
    return new SWTWindowPeer(this, target);
  }

  @Override
  public DialogPeer createDialog(Dialog target) {
    return new SWTDialogPeer(this, target);
  }

  @Override
  public FileDialogPeer createFileDialog(FileDialog target) {
    return new SWTFileDialogPeer(this, target);
  }

  @Override
  public FramePeer createFrame(Frame target) {
    return new SWTFramePeer(this, target);
  }

  @Override
  public RobotPeer createRobot(Robot target, GraphicsDevice screen) {
    // TODO Manage Robot
    return null;
  }

  @Override
  public Clipboard getSystemClipboard()  {
    // TODO Manage Clipboard
    return null;
  }

  public Clipboard getSystemSelection() throws HeadlessException {
    // TODO Manage Clipboard
    return null;
  }

  @Override
  public DragSourceContextPeer createDragSourceContextPeer(DragGestureEvent dge)
      throws InvalidDnDOperationException {
    // TODO Manage DnD
    return null;
  }

  @Override
  public boolean isTraySupported() {
    return invokeInSWTDisplayThread(new Callable<Boolean>() {
      public Boolean call() {
        return getSWTDisplay().getSystemTray() != null;
      }
    });
  }

  @Override
  public SystemTrayPeer createSystemTray(SystemTray target) {
    return new SWTTrayPeer(this, target);
  }

  @Override
  public TrayIconPeer createTrayIcon(TrayIcon target) {
    return new SWTTrayIconPeer(this, target);
  }

  @Override
  public boolean isDesktopSupported() {
    return true;
  }
  
  @Override
  protected DesktopPeer createDesktopPeer(Desktop target) {
    return new SWTDesktopPeer(this, target);
  }

  @Override
  public EventQueue getSystemEventQueueImpl() {
    return super.getSystemEventQueueImpl();
  }

  @Override
  public ColorModel getColorModel() throws HeadlessException {
    return ColorModel.getRGBdefault ();
  }

  @Override
  public String [] getFontList() {
    return invokeInSWTDisplayThread(new Callable<String []>() {
      public String [] call() {
        FontData [] fontsData = getSWTDisplay().getFontList(null, true);
        Set<String> fontsSet = new HashSet<String> ();
        for (FontData fontData : fontsData) {
          fontsSet.add(fontData.getName());
        }
        String [] fonts = fontsSet.toArray(new String [fontsSet.size()]);
        Arrays.sort(fonts);
        return fonts;
      }
    });
  }

  @Override
  public FontMetrics getFontMetrics(final Font font) {
    return invokeInSWTDisplayThread(new Callable<FontMetrics>() {
      public FontMetrics call() {
        if (swtCachedGc == null) {
          swtCachedGc = new GC(getSWTDisplay());
        }
        final org.eclipse.swt.graphics.Font swtFont;
        if (font == null) {
          swtFont = getSWTDisplay().getSystemFont();
        } else {
          swtFont = SWTFonts.getInstance(getSWTDisplay()).getFont(font);
        }
        swtCachedGc.setFont(swtFont);
        org.eclipse.swt.graphics.FontMetrics swtFontMetrics = swtCachedGc.getFontMetrics();
        final int ascent = swtFontMetrics.getAscent();
        final int descent = swtFontMetrics.getDescent();
        final int leading = swtFontMetrics.getLeading();
        
        // Return a new font metrics object 
        return new FontMetrics(font) {
          @Override
          public int getAscent() {
            return ascent;
          }
          
          @Override
          public int getLeading() {
            return leading;
          }
          
          @Override
          public int getDescent() {
            return descent;
          }
          
          @Override
          public int charWidth(final char ch) {
            return invokeInSWTDisplayThread(new Callable<Integer>() {
              public Integer call() {
                swtCachedGc.setFont(swtFont);
                return swtCachedGc.textExtent(String.valueOf(ch)).x;
              }
            });
          }
          
          @Override
          public int charsWidth(final char [] data, final int off, final int len) {
            return invokeInSWTDisplayThread(new Callable<Integer>() {
              public Integer call() {
                swtCachedGc.setFont(swtFont);
                return swtCachedGc.textExtent(new String(data, off, len)).x;
              }
            });
          }
        };
      }
    });
  }

  @Override
  public FontPeer getFontPeer(String name, int style) {
    // TODO Need FontPeer ?
    return null;
  }

  @Override
  public PrintJob getPrintJob(Frame frame, String jobtitle,
                              Properties props) {
    // TODO Manage printing
    return null;
  }

  @Override
  public int getScreenResolution() throws HeadlessException {
    return invokeInSWTDisplayThread(new Callable<Integer>() {
      public Integer call() {
        return getSWTDisplay().getDPI().x;
      }
    });
  }

  @Override
  protected int getScreenHeight() {
    return invokeInSWTDisplayThread(new Callable<Integer>() {
      public Integer call() {
        return getSWTDisplay().getBounds().height;
      }
    });
  }

  @Override
  protected int getScreenWidth() {
    return invokeInSWTDisplayThread(new Callable<Integer>() {
      public Integer call() {
        return getSWTDisplay().getBounds().width;
      }
    });
  }

  @Override
  public Insets getScreenInsets(GraphicsConfiguration gc) {
    return invokeInSWTDisplayThread(new Callable<Insets>() {
      public Insets call() {
        org.eclipse.swt.graphics.Rectangle bounds = getSWTDisplay().getBounds();
        org.eclipse.swt.graphics.Rectangle clientArea = getSWTDisplay().getClientArea();
        return new Insets(clientArea.y, clientArea.x, 
            bounds.height - clientArea.height - clientArea.y, 
            bounds.width - clientArea.width - clientArea.x);
      }
    });
  }

  @Override
  public int getMenuShortcutKeyMask() {
    if (isMacOSX()) {
      return KeyEvent.META_MASK;
    } else {
      return KeyEvent.CTRL_MASK;
    }
  }

  public Cursor createCustomCursor(Image cursor, Point hotSpot, String name) {
    // TODO Create custom cursor
    return super.createCustomCursor(cursor, hotSpot, name);
  }

  public Dimension getBestCursorSize(int preferredWidth,
                                     int preferredHeight) {
    return isWindows() ? new Dimension(16, 16) : new Dimension(32, 32);
  }

  boolean isMacOSX() {
    String osName = System.getProperty("os.name");
    return osName.startsWith("Mac") || osName.startsWith("Darwin");
  }
  
  boolean isWindows() {
    String osName = System.getProperty("os.name");
    return osName.startsWith("Windows");
  }
  
  @Override
  public Map<TextAttribute, ?> mapInputMethodHighlight(InputMethodHighlight highlight) {
    // TODO Manage InputMethodHightlight
    return null;
  }

  public InputMethodDescriptor getInputMethodAdapterDescriptor()  {
    // TODO Manage InputMethodDescriptor
    return null;
  }

  @Override
  public void sync() {
    // TODO Any need to sync ?
  }

  @Override
  public void setDynamicLayout(boolean dynamic) {
    this.dynamicLayout = dynamic;
  }

  @Override
  public boolean isDynamicLayoutActive() throws HeadlessException {
    return this.dynamicLayout;
  }

  @Override
  public boolean isFrameStateSupported(int state) {
    return true;
  }

  @Override
  public boolean isModalityTypeSupported(Dialog.ModalityType modalityType) {
    // TODO Implement modality
    return false;
  }

  @Override
  public boolean isModalExclusionTypeSupported(Dialog.ModalExclusionType exclusionType) {
    // TODO Implement modality
    return false;
  }
  
  @Override
  public boolean getLockingKeyState(int keyCode) {
    // TODO Implement locking key state
    if (!(keyCode == KeyEvent.VK_CAPS_LOCK
          || keyCode == KeyEvent.VK_NUM_LOCK
          || keyCode == KeyEvent.VK_SCROLL_LOCK || keyCode == KeyEvent.VK_KANA_LOCK)) {
      return super.getLockingKeyState(keyCode);
    }
    return super.getLockingKeyState(keyCode);
  }

  @Override
  public void setLockingKeyState(int keyCode, boolean on) {
    // TODO Implement locking key state
    if (!(keyCode == KeyEvent.VK_CAPS_LOCK
          || keyCode == KeyEvent.VK_NUM_LOCK
          || keyCode == KeyEvent.VK_SCROLL_LOCK || keyCode == KeyEvent.VK_KANA_LOCK)) {
      super.setLockingKeyState(keyCode, on);
    }
    super.setLockingKeyState(keyCode, on);
  }

  @Override
  protected void loadSystemColors(int[] systemColors) {
    // TODO Load system colors
  }
  
  @Override
  protected boolean syncNativeQueue() {
    // TODO Any need to sync ?
    return false;
  }

  @Override
  public void grab(Window arg0) {
    // TODO Any need to grab ?    
  }

  @Override
  public void ungrab(Window arg0) {
    // TODO Any need to grab ?    
  }
}
